==> 0.4.zip: kbf/src/http/client.cpp <==
#include "kbf/http/client.h"

#include <freertos/task.h>
#include <esp_log.h>

#include "kbf/macros.h"

using namespace kbf;
using std::make_shared;
using std::string;
using std::shared_ptr;

http::Client::Client(int timeoutMs, bool async) : timeoutMs(timeoutMs), async(async), buffer() {
    ESP_LOGD(TAG, "Client()");
    init();
}

void http::Client::init() {
    ESP_LOGD(TAG, "init()");
    esp_http_client_config_t config{};
    config.event_handler         = handleHttpEvent;
    config.user_data             = this;
    config.host                  = "localhost";
    config.path                  = "/";
    config.disable_auto_redirect = false; // do not set to true, IDF bug :(
    config.max_redirection_count = 0;
    config.timeout_ms            = timeoutMs;
    config.is_async              = async;
    handle = esp_http_client_init(&config);
}

http::Client::~Client() {
    ESP_LOGD(TAG, "~Client()");
    CHECK(esp_http_client_cleanup(handle));
}

std::shared_ptr<kbf::http::Response>
http::Client::performRequest(const kbf::http::Method method, const std::string &url, const nlohmann::json *postData) {
    ESP_LOGD(TAG, "performRequest(); method: %d, url: %s", method, url.c_str());

    if (running) {
        ESP_LOGE(TAG, "request already in progress");
        return nullptr;
    }
    running = true;

    response = shared_ptr<Response>(new Response()); // TODO why doesn't make_shared work here? :/
    buffer.clear();
    string body;

    if (method == GET) {
        esp_http_client_set_method(handle, HTTP_METHOD_GET);
    } else if (method == POST) {
        body = postData ? postData->dump() : "";
        ESP_LOGV(TAG, "request body:");
        ESP_LOG_BUFFER_HEXDUMP(TAG, body.c_str(), body.size(), ESP_LOG_VERBOSE);
        esp_http_client_set_method(handle, HTTP_METHOD_POST);
        esp_http_client_set_header(handle, "Content-Type", "application/json");
        esp_http_client_set_post_field(handle, body.c_str(), body.length());
    } else {
        ESP_LOGW(TAG, "unhandled method: %d", method);
    }

    esp_http_client_set_url(handle, url.c_str());
    auto err = esp_http_client_perform(handle);

    if (err == ESP_ERR_HTTP_MAX_REDIRECT) {
        // TODO this seems get triggered on HTTP 4XX
        ESP_LOGW(TAG, "reached redirect limit");
        updateResponse(*this);
        err = ESP_OK;
    }

    if (err == ESP_ERR_HTTP_EAGAIN) {
        ESP_LOGD(TAG, "async request in progress, returning nullptr");
        return nullptr;
    } else if (err == ESP_OK) {
        ESP_LOGD(TAG, "success");
        return response;
    } else if (err == ESP_ERR_HTTP_FETCH_HEADER && !retry) {
        // TODO this will happen on timeout as well; perhaps set a retry limit in ctor / get / post?
        ESP_LOGW(TAG, "fetching headers failed, maybe the connection was dropped? retrying...");
        retry = true;
        esp_http_client_cleanup(handle);
        init();
        return get(url);
    } else {
        ESP_LOGE(TAG, "HTTP GET failed: %s", esp_err_to_name(err));
        if (onError) onError(*this);
        return nullptr;
    }
}

esp_err_t http::Client::handleHttpEvent(esp_http_client_event_t *event) {
    auto client = (Client *) event->user_data;

    switch (event->event_id) {
        case HTTP_EVENT_ERROR:
            ESP_LOGW(TAG, "fixme: unhandled error event: HTTP_EVENT_ERROR");
            if (client->onError) client->onError(*client);
            client->running = false;
            client->retry   = false;
            break;
        case HTTP_EVENT_ON_CONNECTED:
            ESP_LOGD(TAG, "connected");
            break;
        case HTTP_EVENT_HEADERS_SENT:
            ESP_LOGD(TAG, "request sent");
            break;
        case HTTP_EVENT_ON_HEADER:
            ESP_LOGD(TAG, "header received: %s: %s", event->header_key, event->header_value);
            client->response->headers[event->header_key] = event->header_value;
            break;
        case HTTP_EVENT_ON_DATA:
            ESP_LOGD(TAG, "data received");
            if (esp_http_client_is_chunked_response(client->handle)) {
                ESP_LOGE(TAG, "received chunked data");
                ABORT("not implemented");
            }
            client->buffer.append((char *) event->data, event->data_len);
            break;
        case HTTP_EVENT_ON_FINISH:
            ESP_LOGD(TAG, "finished");
            updateResponse(*client);
            break;
        case HTTP_EVENT_DISCONNECTED:
            ESP_LOGW(TAG, "disconnected");
            client->running = false;
            break;
    }

    return ESP_OK;
}

void http::Client::updateResponse(http::Client &client) {
    client.response->status = esp_http_client_get_status_code(client.handle);
    client.response->body   = client.buffer;

    ESP_LOGD(TAG, "HTTP status: %d, Content-Length: %s", client.response->status,
             client.response->headers["Content-Length"].c_str());
    ESP_LOG_BUFFER_HEXDUMP(TAG, client.response->body.data(), client.response->body.length(), ESP_LOG_VERBOSE);

    if (client.onSuccess) { client.onSuccess(client, *client.response); }
    client.running = false;
    client.retry   = false;
}

shared_ptr<http::Response> http::Client::get(const string &url) {
    ESP_LOGD(TAG, "get(%s)", url.c_str());
    return performRequest(GET, url, nullptr);
}

shared_ptr<http::Response> http::Client::post(const string &url, const nlohmann::json &data) {
    ESP_LOGD(TAG, "post(%s)", url.c_str());
    return performRequest(POST, url, &data);
}

==> 0.5.zip: kbf/src/http/client.cpp <==
#include "kbf/http/client.h"

#include <freertos/task.h>
#include <esp_log.h>

#include "kbf/macros.h"

using namespace kbf;
using std::make_shared;
using std::string;
using std::shared_ptr;
using std::optional;
using std::nullopt;
using std::map;

http::Client::Client(int timeoutMs, bool async) : timeoutMs(timeoutMs), async(async), buffer() {
    ESP_LOGD(TAG, "Client()");
    init();
}

void http::Client::init() {
    ESP_LOGD(TAG, "init()");
    esp_http_client_config_t config{};
    config.event_handler         = handleHttpEvent;
    config.user_data             = this;
    config.host                  = "localhost";
    config.path                  = "/";
    config.disable_auto_redirect = false; // do not set to true, IDF bug :(
    config.max_redirection_count = 0;
    config.timeout_ms            = timeoutMs;
    config.is_async              = async;
    handle = esp_http_client_init(&config);
}

http::Client::~Client() {
    ESP_LOGD(TAG, "~Client()");
    CHECK(esp_http_client_cleanup(handle));
}

std::shared_ptr<kbf::http::Response>
http::Client::performRequest(const kbf::http::Method method, const std::string &url, const nlohmann::json *postData,
                             const optional<const map<string, string>> &headers) {
    ESP_LOGD(TAG, "performRequest(); method: %d, url: %s", method, url.c_str());

    if (running) {
        ESP_LOGE(TAG, "request already in progress");
        return nullptr;
    }
    running = true;

    response = shared_ptr<Response>(new Response()); // TODO why doesn't make_shared work here? :/
    buffer.clear();
    string body;

    if (method == GET) {
        esp_http_client_set_method(handle, HTTP_METHOD_GET);
    } else if (method == POST) {
        body = postData ? postData->dump() : "";
        ESP_LOGV(TAG, "request body:");
        ESP_LOG_BUFFER_HEXDUMP(TAG, body.c_str(), body.size(), ESP_LOG_VERBOSE);
        esp_http_client_set_method(handle, HTTP_METHOD_POST);
        esp_http_client_set_header(handle, "Content-Type", "application/json");
        esp_http_client_set_post_field(handle, body.c_str(), body.length());
    } else {
        ESP_LOGW(TAG, "unhandled method: %d", method);
    }

    if (headers) {
        ESP_LOGD(TAG, "adding headers");
        for (auto const &[key, value] : *headers) {
            ESP_LOGD(TAG, "  %s: %s", key.c_str(), value.c_str());
            esp_http_client_set_header(handle, key.c_str(), value.c_str());
        }
    }

    esp_http_client_set_url(handle, url.c_str());
    auto err = esp_http_client_perform(handle);

    if (err == ESP_ERR_HTTP_MAX_REDIRECT) {
        // TODO this seems get triggered on HTTP 4XX
        ESP_LOGW(TAG, "reached redirect limit");
        updateResponse(*this);
        err = ESP_OK;
    }

    if (err == ESP_ERR_HTTP_EAGAIN) {
        ESP_LOGD(TAG, "async request in progress, returning nullptr");
        return nullptr;
    } else if (err == ESP_OK) {
        ESP_LOGD(TAG, "success");
        return response;
    } else if (err == ESP_ERR_HTTP_FETCH_HEADER && !retry) {
        // TODO this will happen on timeout as well; perhaps set a retry limit in ctor / get / post?
        ESP_LOGW(TAG, "fetching headers failed, maybe the connection was dropped? retrying...");
        retry = true;
        esp_http_client_cleanup(handle);
        init();
        return get(url);
    } else {
        ESP_LOGE(TAG, "HTTP GET failed: %s", esp_err_to_name(err));
        if (onError) onError(*this);
        return nullptr;
    }
}

esp_err_t http::Client::handleHttpEvent(esp_http_client_event_t *event) {
    auto client = (Client *) event->user_data;

    switch (event->event_id) {
        case HTTP_EVENT_ERROR:
            ESP_LOGW(TAG, "fixme: unhandled error event: HTTP_EVENT_ERROR");
            if (client->onError) client->onError(*client);
            client->running = false;
            client->retry   = false;
            break;
        case HTTP_EVENT_ON_CONNECTED:
            ESP_LOGD(TAG, "connected");
            break;
        case HTTP_EVENT_HEADERS_SENT:
            ESP_LOGD(TAG, "request sent");
            break;
        case HTTP_EVENT_ON_HEADER:
            ESP_LOGD(TAG, "header received: %s: %s", event->header_key, event->header_value);
            client->response->headers[event->header_key] = event->header_value;
            break;
        case HTTP_EVENT_ON_DATA:
            ESP_LOGD(TAG, "data received");
            if (esp_http_client_is_chunked_response(client->handle)) {
                ESP_LOGE(TAG, "received chunked data");
                ABORT("not implemented");
            }
            client->buffer.append((char *) event->data, event->data_len);
            break;
        case HTTP_EVENT_ON_FINISH:
            ESP_LOGD(TAG, "finished");
            updateResponse(*client);
            break;
        case HTTP_EVENT_DISCONNECTED:
            ESP_LOGW(TAG, "disconnected");
            client->running = false;
            break;
    }

    return ESP_OK;
}

void http::Client::updateResponse(http::Client &client) {
    client.response->status = esp_http_client_get_status_code(client.handle);
    client.response->body   = client.buffer;

    ESP_LOGD(TAG, "HTTP status: %d, Content-Length: %s", client.response->status,
             client.response->headers["Content-Length"].c_str());
    ESP_LOG_BUFFER_HEXDUMP(TAG, client.response->body.data(), client.response->body.length(), ESP_LOG_VERBOSE);

    if (client.onSuccess) { client.onSuccess(client, *client.response); }
    client.running = false;
    client.retry   = false;
}

shared_ptr<http::Response>
http::Client::get(const string &url, const optional<const map<string, string>> &headers) {
    ESP_LOGD(TAG, "get(%s)", url.c_str());
    return performRequest(GET, url, nullptr, headers);
}

shared_ptr<http::Response>
http::Client::post(const string &url, const nlohmann::json &data, const optional<const map<string, string>> &headers) {
    ESP_LOGD(TAG, "post(%s)", url.c_str());
    return performRequest(POST, url, &data, headers);
}
